/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 946 $ $LastChangedDate: 2002-09-12 22:50:25 +0200 (Thu, 12 Sep 2002) $
 *
 * Util - Various utility functions: diagnostics, compilation flags, etc.
 *
 * Maybe we should just leave types in Ast and move functions here?
 */
 
open Ast;
open Util;

/*** Flags. ***/
compiler_flags flags;

section init 
{
	flags = {
		w_error = false,
		time_report = false,
		force_colours = false,
		auto_colours = false,
	
		// these are set by Main anyway
		w_shadow_local = false,
		w_shadow_label = false,
		w_shadow_global = false,
		w_shadow_xmodule = false,
		w_dead = false,
		w_labeled = false,
		w_pattern = false,
		w_capitalization = false,
		w_ignored_value = false,
		w_ignored_retval = false,
		w_unused_parm = false,
		w_unused_variable = false,
		w_unused_patvar = false,
		w_unused_label = false,
		w_unused_global = false,
		w_unused_open = false,
		verbose_errors = false
	};
}

/*** Diagnostics. ***/
<location>Stack::t loc_stack;
bool bailout;
location last_location;

string warn_colour;
string error_colour;
string ice_colour;
string no_colour;

void init_colours()
{
    bool state = flags.force_colours;

    if (!state && flags.auto_colours)
	state = Sys::isatty(Stdio::fileno(Stdio::stderr));

    if (state) {
	// heh, rasta ;)
	warn_colour = "\e[32;1m";
	error_colour = "\e[33;1m";
	ice_colour = "\e[31;1m";
	no_colour = "\e[0m";
    } else {
	ice_colour = warn_colour = no_colour = error_colour = "";
    }
}

section init 
{
	loc_stack = Stack::create();
	last_location = {file = "unknown", 
			 first_line = 0,
			 first_column = 0,
			 last_line = 0,
			 last_column = 0};
	bailout = false;

	ice_colour = warn_colour = no_colour = error_colour = "";
}

void message(location loc, string msg)
{
    string beg = loc.file + ":" + 
		 itoa(loc.first_line) + ":" +
		 itoa(loc.first_column);

    if (loc.last_line != 0 &&
	(loc.last_line != loc.first_line ||
	 (loc.last_column != 0 && loc.last_column != loc.first_column)))
	    beg = beg + "-" + itoa(loc.last_line) + ":" + 
		  itoa(loc.last_column);

    puts_err(beg + ": " + msg + "\n");
}

void warnl(location loc, string msg)
{
    message(loc, "warning: " + warn_colour + msg + no_colour);
    if (flags.w_error)
	    bailout = true;
}

void verb_warnl(location loc, string msg)
{
    message(loc, warn_colour + "(" + msg + ")" + no_colour);
    if (flags.w_error)
	    bailout = true;
}

void errorl(location loc, string msg)
{
    message(loc, "error: " + error_colour + msg + no_colour);
    bailout = true;
}

void verb_errorl(location loc, string msg)
{
    if (flags.verbose_errors)
	message(loc, error_colour + "(" + msg + ")" + no_colour);
    bailout = true;
}


/* Location stack. */
void push_location(location loc)
{
	last_location = loc;
	Stack::push(loc_stack, loc);
}

void pop_location()
{
	Stack::pop(loc_stack);
}

location top_location()
{
	if (Stack::empty(loc_stack))
		return last_location;
	else
		return Stack::top(loc_stack);
}

void warn(string msg)
{
	warnl(top_location(), msg);
}

void error(string msg)
{
	errorl(top_location(), msg);
}

void verb_warn(string msg)
{
	verb_warnl(top_location(), msg);
}

void verb_error(string msg)
{
	verb_errorl(top_location(), msg);
}

void maybe_choke()
{
	if (bailout == true)
		raise Bail_out;
}

'a locate(*('a ('b)) f, <'b>located x)
{
	'a r;

	push_location(x.loc);
	try {
		r = f(x.a);
	} finally {
		pop_location();
	}

	return r;
}

'a ice(string msg)
{
	if (bailout == true) {
		error("confused by earlier errors, bailing out [" + msg + "]");
		raise Bail_out;
	}
	
	error(ice_colour + "internal compiler error: " + msg);
	return Sys::abort();
}


/*** General utils. ***/
'a get_binding(<'a>bindable sym)
{
	switch sym {
	case Symbol(x):
		return x;
	case Simple_name(n):
		return ice("get_binding(): symbol '" + n + "' not bound");
	case Qualified_name(m, n):
		return ice("get_binding(): qualified symbol '" + 
			   m + "::" + n + "' not bound");
	}
}

void dbg(string s)
{
	puts(s);
}

void dbgl(string s)
{
	message(top_location(), "DEBUG: " + s);
}

bool same_symbols(<'a>symbol s1, <'a>symbol s2)
{
	return s1.module == s2.module && s1.name == s2.name;
}

// Unique name generator

struct ung_counter {
	int cnt;
}

<ung_counter>Hashtbl::t ung_counters;

section init
{
	ung_counters = Hashtbl::create(13);
}

string genname(string prefix)
{
	int cnt;
	
	if (Hashtbl::mem(ung_counters, prefix)) {
		let c = Hashtbl::find(ung_counters, prefix) in
		cnt = c.cnt++;
	} else {
		cnt = 1;
		Hashtbl::add(ung_counters, prefix, ({ cnt = 2 }));
	}

	return prefix + "$" + itoa(cnt);
}

// Ast manipulation

string simple_name(<<'a>symbol>bindable s)
{
    switch s {
    case Simple_name(x):
	return x;
	
    case Qualified_name(_, x):
	return x;
	
    case Symbol(x):
	return x.name;
    }
}

string symbol_name(<'a>symbol x)
{
    if (x.module != "")
	return x.module + "::" + x.name;
    else
	return x.name;
}

string qualified_name(<<'a>symbol>bindable s)
{
    switch s {
    case Simple_name(x):
	return x;
	
    case Qualified_name(m, n):
	return m + "::" + n;
	
    case Symbol(x):
	return symbol_name(x);
    }
}

<<'a>symbol>bindable rev_qualified_name(string name)
{
    <string>list l = String::split("::", name);
    
    switch l {
    case Cons(m, Cons(n, Nil)):
	return Qualified_name(m, n);
    case Cons(n, Nil):
	return Simple_name(n);
    case _:
	raise Invalid_argument("Ast::rev_qualified_name");
    }
}

void set_property(var_symbol s, var_symbol_property p)
{
    if (Hacks::union_tag(p) >= 32)
	ice("Util::set_property(): fixme: too much properties");
    
    s.a.properties = s.a.properties | (1 << Hacks::union_tag(p));
}

bool has_property(var_symbol s, var_symbol_property p)
{
    if (Hacks::union_tag(p) >= 32)
	ice("Util::has_property(): fixme: too much properties");
    
    return (s.a.properties & (1 << Hacks::union_tag(p))) != 0;
}

ty type_map(*(ty (ty)) f, ty t)
{
    ty self(ty tt)
    {
	return type_map(f, tt);
    }

    <'a>field self_fld(<'a>field fld)
    {
	ty tt;

	push_location(fld.fld_loc);
	try {
	    tt = self(fld.fld_type);
	} finally {
	    pop_location();
	}
	
	return ({fld_type = tt,
	     fld_symbol = fld.fld_symbol,
	     fld_loc = fld.fld_loc});
    }
    
    t = f(t);
    
    switch t {
    case Tuple(lst):
	return Tuple(List::map(self, lst));
    
    case Func(ft):
	*(ty, <id>option) map_fnc_arg(*(ty, <id>option) p)
	{
	    let (tt, nn) = p in
	    return (self(tt), nn);
	}
	
	return Func(({ret_type = self(ft.ret_type),
		      parms = List::map(map_fnc_arg, ft.parms),
		      tyvars = ft.tyvars}));
    
    case Struct(flds):
	return Struct(List::map(self_fld, flds));
	
    case Opt_struct(flds):
	return Opt_struct(List::map(self_fld, flds));
	
    case Union(flds):
	return Union(List::map(self_fld, flds));
    
    case Named_type(nt):
	return Named_type(({parms = List::map(self, nt.parms),
		   	   tydef = nt.tydef}));

    case X_tyvar(_):
	return t;   // ???

    case _:
	return t;
    }
}

ty copy_type(ty t)
{
    ty f(ty tt)
    {
	switch tt {
	case X_tyvar(x):
	    _ xx = {t = None, name = x.name + "c", flag = x.flag};
	    switch x.t {
	    case Some(y):
		xx.t = Some(copy_type(y));
	    case None:
		skip;
	    }
	    return X_tyvar(xx);
	case _:
	    return tt;
	}
    }

    return type_map(f, t);
}

void type_iter(*(void (ty)) f, ty t)
{
    void self(ty tt)
    {
	type_iter(f, tt);
    }

    void self_fld(<'a>field fld)
    {
	push_location(fld.fld_loc);
	try {
	    self(fld.fld_type);
	} finally {
	    pop_location();
	}
    }
    
    f(t);
    
    switch t {
    case Tuple(lst):
	List::iter(self, lst);
    
    case Func(ft):
	void map_fnc_arg(*(ty, <id>option) p)
	{
	    let (tt, _) = p in
	    self(tt);
	}
	
	self(ft.ret_type);
	List::iter(map_fnc_arg, ft.parms);
    
    case Struct(flds):
	List::iter(self_fld, flds);
	
    case Opt_struct(flds):
	List::iter(self_fld, flds);
	
    case Union(flds):
	List::iter(self_fld, flds);
    
    case Named_type(nt):
	List::iter(self, nt.parms);
	
    case X_tyvar(xt):
	switch xt.t {
	case Some(tt):
	    self(tt);
	case None:
	    skip;
	}
    
    case _:
	skip;
    }
}

string make_string(location l, Iconv::cstring ptr, int len)
{
    try {
	return Iconv::string_of_utf8_len(ptr, len);
    } with {
    case Invalid_multibyte_sequence:
	errorl(l, "invalid utf-8 seqence in string");
    case Incomplete_multibyte_sequence:
	errorl(l, "incomplete utf-8 seqence in string");
    }

    return "";
}

string current_module;
bool doing_iface;

section init
{
    current_module = "?";
    doing_iface = false;
}

// vim:sts=4:sw=4
