/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 946 $ $LastChangedDate: 2002-09-12 22:50:25 +0200 (Thu, 12 Sep 2002) $
 *
 * Util - Various utility functions: diagnostics, compilation flags, etc.
 */
 
open Ast;

/* Prints message and sets bailout flag.  The flag is tested by maybe_choke()
   routine.  */
void errorl(location loc, string msg);

/* Prints error message if verbose errors flag is set.  */
void verb_errorl(location loc, string msg);

/* Prints message prepended with 'warning: ', sets bailout flag is -Werror
   was specified in command line.  */
void warnl(location loc, string msg);

/* Prints warning message if verbose errors flag is set.  */
void verb_warnl(location loc, string msg);

/* If bailout flag is set raise Exit.  */
void maybe_choke();

/** Location stack. 

    Whenever we start processing some language construct with location
    included in Ast, we push its location, on stack.  After we are finish
    with it, we pop it.  This allows various routines just to call
    error()/warning() functions without any location, and have the location
    automagically prepended.  */

/* Push location on the stack.  */
void push_location(location loc);

/* Pop it.  */
void pop_location();

/* Return topmost location of the stack. If stack is empty we return last
   location pushed, no matter if it was popped.  This is used in constructions
   of artificial Ast's.  */
location top_location();

/* Print warning message in topmost location context.  */
void warn(string msg);
void verb_warn(string msg);

/* Print error message in topmost location context.  */
void error(string msg);
void verb_error(string msg);

/* Run [f(x.a)] with [x.loc] pushed before and popped after.  */
'a locate(*('a ('b)) f, <'b>located x);

/* Prints error message and does abort().  ICE == Internal Compiler Error.  
   It uses top_location() so user can figure out how to code it around.
   If bailout flag is set, then we print "confused by earlier errors,
   bailing out", otherwise "internal compiler error".  */
'a ice(string msg);

/* Returning <'a>symbol for given binding.  If not bound -- ICE.  */
'a get_binding(<'a>bindable sym);

/* Compiler flags.  */
struct compiler_flags {
	bool w_error;
	bool w_shadow_local;
	bool w_shadow_label;
	bool w_shadow_global;
	bool w_shadow_xmodule;
	bool w_dead;
	bool w_labeled;
	bool w_pattern;
	bool w_capitalization;
	bool w_ignored_value;
	bool w_ignored_retval;
	bool w_unused_parm;
	bool w_unused_variable;
	bool w_unused_patvar;
	bool w_unused_label;
	bool w_unused_global;
	bool w_unused_open;
	bool verbose_errors;
	bool time_report;
	bool force_colours;
	bool auto_colours;
}

/* Current compiler settings.  */
compiler_flags flags;

/** Debugging utilities.  */

/* Same as [print_string(m)]. */
void dbg(string m);

/* Like [dbg()] but also prepend line number and append newline.  */
void dbgl(string s);

/* Return [true] iff two [s1] and [s2] are the same symbol.  */
bool same_symbols(<'a>symbol s1, <'a>symbol s2);

/* Return unique name starting with [prefix].  */
string genname(string prefix);

/* Apply [f] to [t], and the recursively to each component type of
   what it returns.  For example [f] return Struct(...), then it's
   later applied to each [fld_type] and such mapped structure is
   returned.  It does not walk down X_tyvar(...) chains.  */
ty type_map(*(ty (ty)) f, ty t);

/* Similar to [type_map()], but [f] can only modify [t] in-place, not return
   new copy of it.  It also differs from [type_map()] in that it walks down
   X_tyvar(...) chains.  */
void type_iter(*(void (ty)) f, ty t);

/* Return fresh copy of [t].  Used in making Ast of [_ x, y, z;] where
   each of [x, y, z] gets it's own copy of weak type.  */
ty copy_type(ty t);

/* Add [p] to [s] properties list.  */
void set_property(var_symbol s, var_symbol_property p);

/* Check whether [s] has property [p].  */
bool has_property(var_symbol s, var_symbol_property p);

/* Return just name component of symbol.  */
string simple_name(<<'a>symbol>bindable s);

/* Return [module::name] form name. If symbol name includes no module -- return
   just [name].  */
string qualified_name(<<'a>symbol>bindable s);

/* Decomposite [module::name] form name into Qualified_name(...) or 
   Simple_name(...).  Used by module serializer.  */
<<'a>symbol>bindable rev_qualified_name(string name);

/* Same as [qualified_name(Symbol(x)].  */
string symbol_name(<'a>symbol x);

/* Raised by routines that cannot return anything because of user error.
   It might be cought if value returned by function in question is not needed
   at some higher level.  If it gets to main, compilation terminates.  */
exception void Choke;

/* Raised when we really want to get out the compiler.  */
exception void Bail_out;

/* Name of currently compiled module.  */
string current_module;

/* [true] iff we are compiling .gi file.  */
bool doing_iface;

/* Initialize colors setting according to flags.  */
void init_colours();

string make_string(location l, Iconv::cstring ptr, int len);
