/**
+ Adding callbacks to RPN calculator

We'll now change execution model of our calculator a bit.  Instead of
using one big switch over operators, we'll make hash table mapping
operator names to functions.  An union of few kinds of functions
will be used.
*/

// just like before
union slot {
	int Number;
	string Variable;
}

<slot>Stack::t stack = Stack::create();
<int>Hashtbl::t variable_bindings = Hashtbl::create(0);

// Raised in case user causes an error.  Argument is human-readable
// error message.
exception string Error;

// Push number on the stack.
void push(int x)
{
	Stack::push(stack, Number(x));
}

// Pop slot off the stack and convert it to number.  Raise [Error]
// if it's impossible.
int pop()
{
	try {
		slot s = Stack::pop(stack);
		
		switch s {
		case Number(x):
			return x;
			
		case Variable(n):
			try {
				return Hashtbl::find(variable_bindings, n);
			} with {
			case Not_found:
				raise Error("no such variable: " + n);
			}
		}
	} with {
	case Invalid_argument(_):
		raise Error("empty stack");
	}
}

// new stuff
union operator {
	*(int (int, int)) Binary;		// {0}
	*(int (int)) Unary;
	*(void ()) Special;
}

<operator>Hashtbl::t ops = Hashtbl::create(0);

void register_operators()
{
	int add(int a, int b) { return a + b; }			// {1}
	int sub(int a, int b) (a - b); 				// {2}

	Hashtbl::add(ops, "-", Binary(sub));		// {3}
	Hashtbl::add(ops, "+", Binary(add));
	Hashtbl::add(ops, "*", 
		Binary(fun int (int a, int b) { return a * b; } )); 	// {4}
	Hashtbl::add(ops, "/", 
		Binary(fun int (int a, int b) (b / a))); 		// {5}

	void print_top()
	{
		int v1 = pop();
		puts(itoa(v1) + "\n");
		push(v1);
	}

	void set()
	{
		int v1 = pop();
		try {
			switch Stack::pop(stack) {
			case Variable(n):
				Hashtbl::add(variable_bindings, n, v1);

			case Number(_):
				raise Error("expected variable, found number");
			}
		} with {
		case Invalid_argument(_):
			raise Error("empty stack");
		}
	}

	Hashtbl::add(ops, "s", Special(set));
	Hashtbl::add(ops, "p", Special(print_top));
	Hashtbl::add(ops, "q", Special(fun void () {raise End_of_file;}));

	Hashtbl::add(ops, "n", Unary(fun int (int x) (-x)));
}

void perform_operation(string cmd)
{
	// {6}
	if (Hashtbl::mem(ops, cmd))
		switch Hashtbl::find(ops, cmd) {
		case Binary(f):
			int v1 = pop(), v2 = pop();
			push(f(v2, v1));
		case Unary(f):
			push(f(pop()));
		case Special(f):
			f();
		}
	else
		try {
			push(atoi(cmd));
		} with {
		case Invalid_argument(_):
			Stack::push(stack, Variable(cmd));
		}
}

void main()
{
	register_operators();

	for (;;)
		try {
			perform_operation(read_string());
		} with {
		case Error(msg):
			puts("Error: " + msg + ".\n");
		case End_of_file:
			puts("Bye.\n");
			Sys::exit(0);
		}
}

section init
{
	main();
}

/**
*(int (int, int)) is type of function taking two integer parameters
and returning one integer.  For example after [int add(int x, int y) 
{ return x + y; }] definition [add] has this type.  Notation of the 
type reassembles function defintion syntax.

{0} Here we define union of three function types, for binary operators,
unary operators and special operators.  In fact we could use [Special]
for everything, since it can pop and push variables on its own, but it
would be inconvinient to do the same thing the beginning and the end of
function defining operator (two [pop()] operations, and one [push()]).
Therefore [Special] is reserved for operators like "p" or "q".

{1} Here we define nested function to perform addition operation.
There is nothing new about this definition.

{2} Here we use alternative syntax for function defintion.  
[(foo);] is the same as [{ return foo; }].  It's often handy for
very simple functions.

{3} Here functional value [sub] is passed to function [Hashtbl:add]
in order to add it to the hash table.  The name of nested or global 
function is used to form its functional value.

{4} Here the funny part begins :-) Functions need not be named in order
to be passed.  They can be defined inline as well.  [fun] keyword is
used to define functional values without naming them.  This kind of
expressions is called lambda expressions.  As you can see syntax is
very similar to regular function definition, except that is starts with
[fun] and there is no function name.

{5} Short form of return statement is also allowed in lambda expressions.

{6} Here we use slightly different model of accessing hash table.  We first
explictly check whether or not given key is in the hashtable, and then
lookup it.  We do not care about any exceptions.
*/
